import tagpro_eu
import unittest


class JsonTestObject(tagpro_eu.JsonObject):
    __fields__ = {
        'foo': int,
        '__bar__': int
    }


class JsonTestObjectNested(tagpro_eu.JsonObject):
    __fields__ = {
        'inner': JsonTestObject
    }


class JsonTestObjectWithList(tagpro_eu.JsonObject):
    __fields__ = {
        'list': tagpro_eu.data.ListOf(JsonTestObject)
    }


class TestJsonObject(unittest.TestCase):
    def get_object(self):
        return JsonTestObject({'foo': 3, 'bar': 5})

    def test_init(self):
        o = self.get_object()
        self.assertEqual(o.foo, 3)

    def test_hidden_field(self):
        o = self.get_object()
        self.assertEqual(o.__bar__, 5)

    def test_strict(self):
        with self.assertRaises(KeyError):
            JsonTestObject({}, strict=True)

        with self.assertRaises(ValueError):
            JsonTestObject({'foo': 'bar', 'bar': 5}, strict=True)

    def test_not_strict(self):
        JsonTestObject({}, strict=False)
        JsonTestObject({'foo': 'bar', 'bar': 5}, strict=False)

    def test_to_dict(self):
        o = self.get_object()
        self.assertEqual(o, JsonTestObject(o.to_dict()))

    def test_equality(self):
        o = self.get_object()
        p = self.get_object()
        q = JsonTestObject({'foo': 3, 'bar': 6})
        self.assertEqual(o, p)
        self.assertNotEqual(o, q)
        self.assertNotEqual(p, q)


class TestJsonObjectNesting(unittest.TestCase):
    def nested_object(self):
        return JsonTestObjectNested({'inner': {'foo': 3}})

    def test_nesting(self):
        o = self.nested_object()
        self.assertEqual(o.inner.foo, 3)

    def test_parent(self):
        o = self.nested_object()
        self.assertEqual(o, o.inner.__parent__)

    def test_to_dict(self):
        o = self.nested_object()
        self.assertEqual(o, JsonTestObjectNested(o.to_dict()))


class TestJsonObjectList(unittest.TestCase):
    def list_object(self):
        return JsonTestObjectWithList({'list': [{'foo': 3}, {'foo': 4}]})

    def test_list(self):
        o = self.list_object()
        self.assertEqual(o.list[0].foo, 3)
        self.assertEqual(o.list[1].foo, 4)

    def test_index(self):
        o = self.list_object()
        self.assertEqual(o.list[0].index, 0)
        self.assertEqual(o.list[1].index, 1)

    def test_parent(self):
        o = self.list_object()
        self.assertEqual(o.list[0].__parent__, o)
        self.assertEqual(o.list[1].__parent__, o)

    def test_to_dict(self):
        o = self.list_object()
        self.assertEqual(o, JsonTestObjectWithList(o.to_dict()))


class TestMatch(unittest.TestCase):
    def test_player_in_team(self):
        match = tagpro_eu.match.Match({
            'players': [
                {'name': 'foo', 'team': 1},
                {'name': 'bar', 'team': 2}
            ],
            'teams': [
                {},
                {}
            ]
        })

        p1, p2 = match.players
        t1, t2 = match.teams

        self.assertNotEqual(p1, p2)
        self.assertNotEqual(t1, t2)

        self.assertIn(p1, t1.players)
        self.assertIn(p2, t2.players)
        self.assertNotIn(p1, t2.players)
        self.assertNotIn(p2, t1.players)

        self.assertEqual(p1.team, t1)
        self.assertEqual(p2.team, t2)



import json

from tagpro_eu.map import Map
from tagpro_eu.match import Match


def load_matches(f, maps=None):
    """
    Read a file containing bulk match data, and yield the matches.
    A bulk file can be downloaded from https://tagpro.eu/?science

    The tagpro.eu bulk file is a JSON object mapping Match objects to their
    MatchId values. The MatchId is added as the property match_id in the
    yielded Match objects.

    If you want, you can supply a maps object (also as downloaded from
    tagpro.eu) to fill map data from. This object can be loaded using the
    load_maps method.

    :param f: a file descriptor to read matches from
    :param maps: the maps object (omit if undesired)
    :returns: the matches contained in the file
    """
    data = json.load(f)
    for k, v in data.items():
        match = Match(v)
        match.match_id = k
        match.map_id = v['mapId']
        if maps is not None:
            match.map = maps[match.map_id]
        yield match


def load_maps(f):
    """
    Read a file and return a maps object to be used in bulk_matches.
    The bulk maps file can be downloaded from https://tagpro.eu/?science

    :param f: a file descriptor to read maps from
    :returns: the maps object
    """
    data = json.load(f)
    return {int(k): Map(v) for k, v in data.items()}



from tagpro_eu import Blob
import base64
import unittest


class TestBlob(unittest.TestCase):
    def test_base64(self):
        data = bytes([0b0, 0b1, 0b10, 0b11, 0b100])
        blob = Blob.from_b64(base64.b64encode(data))
        self.assertEqual(blob.data, data)

    def test_read_bool(self):
        data = bytes([0b01001011])
        blob = Blob(data)
        self.assertEqual(blob.read_bool(), 0)
        self.assertEqual(blob.read_bool(), 1)
        self.assertEqual(blob.read_bool(), 0)
        self.assertEqual(blob.read_bool(), 0)
        self.assertEqual(blob.read_bool(), 1)
        self.assertEqual(blob.read_bool(), 0)
        self.assertEqual(blob.read_bool(), 1)
        self.assertEqual(blob.read_bool(), 1)

    def test_read_fixed(self):
        data = bytes([0b11010101, 0b11110010, 0b10010101])
        blob = Blob(data)
        self.assertEqual(blob.read_fixed(1), 0b1)
        self.assertEqual(blob.read_fixed(3), 0b101)
        self.assertEqual(blob.read_fixed(2), 0b01)
        self.assertEqual(blob.read_fixed(4), 0b0111)
        self.assertEqual(blob.read_fixed(8), 0b11001010)
        self.assertEqual(blob.read_fixed(6), 0b010101)

    def test_read_tally(self):
        data = bytes([0b11101011, 0b11001010])
        blob = Blob(data)
        self.assertEqual(blob.read_tally(), 3)
        self.assertEqual(blob.read_tally(), 1)
        self.assertEqual(blob.read_tally(), 4)
        self.assertEqual(blob.read_tally(), 0)
        self.assertEqual(blob.read_tally(), 1)
        self.assertEqual(blob.read_tally(), 1)

    def test_reset(self):
        data = bytes([0b10000000])
        blob = Blob(data)
        self.assertEqual(blob.read_bool(), 1)
        blob.reset()
        self.assertEqual(blob.pos, 0)
        self.assertEqual(blob.read_bool(), 1)



from collections import defaultdict
import heapq

from tagpro_eu.blob import Blob
from tagpro_eu.constants import Flag, Flair, Powerup, Team
from tagpro_eu.data import JsonObject
from tagpro_eu.util import Time


class Player(JsonObject):
    """
    Represents a player object in tagpro.eu match files.
    The exact format can be found on https://tagpro.eu/?science
    """
    __fields__ = {
        'auth': bool,
        'name': str,
        'flair': Flair,
        'degree': int,
        'score': int,
        'points': int,    # rank points
        '__team__': int,  # at start of match; 1 = red, 2 = blue, 0 = join late
        'events': Blob.from_b64,
    }

    def __init__(self, data, strict=False):
        super().__init__(data, strict=strict)

        self.__stats__ = None
        self.__caps_for__ = None
        self.__caps_against__ = None

    @property
    def caps_for(self):
        """
        Return the number of captures done by the player's team.
        """
        if self.__caps_for__ is None:
            self.__parent__.__compute_cap_diff__()
        return self.__caps_for__

    @property
    def caps_against(self):
        """
        Return the number of captures done by the player's enemy's team.
        """
        if self.__caps_against__ is None:
            self.__parent__.__compute_cap_diff__()
        return self.__caps_against__

    @property
    def cap_diff(self):
        """
        Return caps_for - caps_against.
        """
        return self.caps_for - self.caps_against

    @property
    def team(self):
        """
        Return the MatchTeam object for the team this player is in.
        """
        return self.__parent__.team(self.__team__)

    @property
    def stats(self):
        """
        Return a PlayerStats object corresponding to the player's stats. This
        is lazy-loaded from the events blob.

        :returns: the PlayerStats object for this player
        """
        if self.__stats__ is None:
            handler = PlayerStats()
            self.parse_events(handler)
            handler.caps_for = self.caps_for
            handler.caps_against = self.caps_against
            self.__stats__ = handler
        return self.__stats__

    def parse_events(self, handler):
        """
        Parse the events blob using the given reader.

        :param handler: the PlayerEventHandler object used to read the events
        """
        team = self.__team__
        duration = self.__parent__.duration

        blob = self.events
        blob.reset()

        time = Time()
        flag = Flag.none
        powers = Powerup.none
        prevent = False
        button = False
        block = False

        if team:
            handler.join(time, team)

        while not blob.end():
            new_team = team
            if blob.read_bool():
                if team == Team.none:
                    new_team = Team(1 + blob.read_bool())
                elif blob.read_bool():
                    new_team = Team.none
                elif team == Team.red:
                    new_team = Team.blue
                elif team == Team.blue:
                    new_team = Team.red

            drop_pop = blob.read_bool()
            returns = blob.read_tally()
            tags = blob.read_tally()
            grab = not flag and blob.read_bool()
            captures = blob.read_tally()

            # wtf
            keep = not drop_pop and new_team and \
                (new_team == team or not team) and \
                (not captures or not flag and
                 not grab or blob.read_bool())

            new_flag = flag
            if grab:
                if keep:
                    new_flag = Flag(1 + blob.read_fixed(2))
                else:
                    new_flag = Flag.temp

            powerups = blob.read_tally()
            pdown = Powerup.none
            pup = Powerup.none

            for p in Powerup.enumerate():
                if powers & p:  # Short-circuit 'and' will NOT work
                    if blob.read_bool():
                        pdown |= p
                elif powerups:
                    if blob.read_bool():
                        pup |= p
                        powerups -= 1

            toggle_prevent = blob.read_bool()
            toggle_button = blob.read_bool()
            toggle_block = blob.read_bool()

            time += 1 + blob.read_footer()

            if not team and new_team:
                team = new_team
                handler.join(time, team)

            for i in range(returns):
                handler.return_(time, flag, powers, team)

            for i in range(tags):
                handler.tag(time, flag, powers, team)

            if grab:
                flag = new_flag
                handler.grab(time, flag, powers, team)

            for i in range(captures):
                if keep or not flag:
                    handler.flagless_capture(time, flag, powers, team)
                else:
                    handler.capture(time, flag, powers, team)
                    flag = Flag.none
                    keep = True

            for p in Powerup.enumerate():
                if pdown & p:
                    powers ^= p
                    handler.powerdown(time, flag, p, powers, team)
                elif pup & p:
                    powers |= p
                    handler.powerup(time, flag, p, powers, team)

            for i in range(powerups):
                handler.duplicate_powerup(time, flag, powers, team)

            if toggle_prevent:
                if prevent:
                    handler.stop_prevent(time, flag, powers, team)
                    prevent = False
                else:
                    handler.start_prevent(time, flag, powers, team)
                    prevent = True

            if toggle_button:
                if button:
                    handler.stop_button(time, flag, powers, team)
                    button = False
                else:
                    handler.start_button(time, flag, powers, team)
                    button = True

            if toggle_block:
                if block:
                    handler.stop_block(time, flag, powers, team)
                    block = False
                else:
                    handler.start_block(time, flag, powers, team)
                    block = True

            if drop_pop:
                if flag:
                    handler.drop(time, flag, powers, team)
                    flag = Flag.none
                else:
                    handler.pop(time, powers, team)

            if new_team != team:
                if not new_team:
                    handler.quit(time, flag, powers, team)
                    powers = Powerup.none
                else:
                    handler.switch(time, flag, powers, new_team)

                flag = Flag.none
                team = new_team

        handler.end(duration, flag, powers, team)

    def __lt__(self, other):
        """
        Necessary for building a heap of events.
        This obviously does not actually work.
        """
        return False

    def __eq__(self, other):
        """
        There's no "safe" way to compare players across matches, as even if a
        player is authorized, their name might have belonged to someone else
        in the past.

        If you analyze public games, I strongly recommend against using == on
        players, and writing your own method instead. For private/competitive
        games, this method should be fine though.
        """
        return other is not None and self.name == other.name

    def __repr__(self):
        return f'Player(name={self.name!r})'


class PlayerEventHandler:
    """
    Event handler for reading a player's events blob.
    """
    def join(self, time, new_team):
        pass

    def quit(self, time, old_flag, old_powers, old_team):
        pass

    def switch(self, time, old_flag, powers, new_team):
        pass

    def grab(self, time, new_flag, powers, team):
        pass

    def capture(self, time, old_flag, powers, team):
        pass

    def flagless_capture(self, time, flag, powers, team):
        pass

    def powerup(self, time, flag, power_up, new_powers, team):
        pass

    def duplicate_powerup(self, time, flag, powers, team):
        pass

    def powerdown(self, time, flag, power_down, new_powers, team):
        pass

    def return_(self, time, flag, powers, team):
        pass

    def tag(self, time, flag, powers, team):
        pass

    def drop(self, time, old_flag, powers, team):
        pass

    def pop(self, time, powers, team):
        pass

    def start_prevent(self, time, flag, powers, team):
        pass

    def stop_prevent(self, time, flag, powers, team):
        pass

    def start_button(self, time, flag, powers, team):
        pass

    def stop_button(self, time, flag, powers, team):
        pass

    def start_block(self, time, flag, powers, team):
        pass

    def stop_block(self, time, flag, powers, team):
        pass

    def end(self, time, flag, powers, team):
        pass


class PlayerStats(PlayerEventHandler):
    """
    Implementation of PlayerEventHandler that accumulates the player's stats.
    """
    def __init__(self):
        self.tags = 0
        self.pops = 0
        self.grabs = 0
        self.drops = 0
        self.hold = Time()
        self.captures = 0
        self.returns = 0
        self.prevent = Time()
        self.button = Time()
        self.block = Time()
        self.pups = defaultdict(int)
        self.pup_time = defaultdict(Time)
        self.time = Time()

        self.ingame_since = -1
        self.hold_since = -1
        self.prevent_since = -1
        self.button_since = -1
        self.block_since = -1
        self.pup_since = defaultdict(lambda: -1)

        # these are supplied by the player object
        self.caps_for = self.caps_against = 0

    @property
    def cap_diff(self):
        return self.caps_for - self.caps_against

    @property
    def pups_total(self):
        """
        Return the total number of powerups picked up by the player.

        :returns: number of powerups
        """
        return sum(self.pups.values())

    def __add__(self, other):
        new = PlayerStats()
        keys = ['tags', 'pops', 'grabs', 'drops', 'hold', 'captures',
                'returns', 'prevent', 'button', 'block', 'time',
                'caps_for', 'caps_against']
        pup_dicts = ['pups', 'pup_time']

        for k in keys:
            setattr(new, k, getattr(self, k) + getattr(other, k))

        for k in pup_dicts:
            for s in self, other:
                for p, n in getattr(s, k).items():
                    getattr(new, k)[p] += n

        return new

    def join(self, time, new_team):
        self.ingame_since = time

    def quit(self, time, old_flag, old_powers, old_team):
        if self.ingame_since >= 0:
            self.time += time - self.ingame_since
            self.ingame_since = -1

    def switch(self, time, old_flag, powers, new_team): pass

    def grab(self, time, new_flag, powers, team):
        self.grabs += 1
        self.hold_since = time

    def capture(self, time, old_flag, powers, team):
        self.captures += 1

        if self.hold_since >= 0:
            self.hold += time - self.hold_since
            self.hold_since = -1

    def flagless_capture(self, time, flag, powers, team):
        self.captures += 1

    def powerup(self, time, flag, power_up, new_powers, team):
        self.pups[power_up] += 1
        self.pup_since[power_up] = time

    def duplicate_powerup(self, time, flag, powers, team):
        self.pups[Powerup.none] += 1

    def powerdown(self, time, flag, power_down, new_powers, team):
        if self.pup_since[power_down] >= 0:
            self.pup_time[power_down] += time - self.pup_since[power_down]
            self.pup_since[power_down] = -1

    def return_(self, time, flag, powers, team):
        self.returns += 1
        self.tags += 1

    def tag(self, time, flag, powers, team):
        self.tags += 1

    def drop(self, time, old_flag, powers, team):
        self.pops += 1
        self.drops += 1

        if self.hold_since >= 0:
            self.hold += time - self.hold_since
            self.hold_since = -1

    def pop(self, time, powers, team):
        self.pops += 1

    def start_prevent(self, time, flag, powers, team):
        self.prevent_since = time

    def stop_prevent(self, time, flag, powers, team):
        if self.prevent_since >= 0:
            self.prevent += time - self.prevent_since
            self.prevent_since = -1

    def start_button(self, time, flag, powers, team):
        self.button_since = time

    def stop_button(self, time, flag, powers, team):
        if self.button_since >= 0:
            self.button += time - self.button_since
            self.button_since = -1

    def start_block(self, time, flag, powers, team):
        self.block_since = time

    def stop_block(self, time, flag, powers, team):
        if self.block_since >= 0:
            self.block += time - self.block_since
            self.block_since = -1

    def end(self, time, flag, powers, team):
        if self.ingame_since >= 0:
            self.time += time - self.ingame_since
            self.ingame_since = -1

        for p in Powerup.enumerate():
            if self.pup_since[p] >= 0:
                self.pup_time[p] += time - self.pup_since[p]
                self.pup_since[p] = -1

        if self.hold_since >= 0:
            self.hold += time - self.hold_since
            self.hold_since = -1

        if self.prevent_since >= 0:
            self.prevent += time - self.prevent_since
            self.prevent_since = -1

        if self.button_since >= 0:
            self.button += time - self.button_since
            self.button_since = -1

        if self.block_since >= 0:
            self.block += time - self.block_since
            self.block_since = -1


class PlayerEventLogger(PlayerEventHandler):
    """
    Implementation of PlayerEventHandler that logs all events to a priority
    queue (MinHeap). Each earliest event can be collected from the queue by
    repeatedly calling heappop from the module heapq.
    """
    def __init__(self, heap=None, player=None):
        """
        Initialize an event logger with the given queue and player object.
        If no queue is given, a new one is made which can be later acquired
        from the heap attribute.

        :param heap: the priority queue to log to
        :param player: the Player object corresponding to this logger
        """
        if heap is None:
            heap = []

        self.heap = heap
        self.player = player

    def get_team_name(self, team):
        """
        Return the name of a team. If a Player object was given in the
        constructor, the actual team name as set in the json object is used.
        Otherwise, the default name (red/blue) is used.

        :param team: the team to get the name of
        :returns: the name of the team
        """
        if self.player is None:
            return str(team)

        return self.player.__parent__.team(team).name

    def join(self, time, new_team):
        heapq.heappush(
            self.heap,
            (time, f'Join team {self.get_team_name(new_team)}', self.player))

    def quit(self, time, old_flag, old_powers, old_team):
        heapq.heappush(
            self.heap,
            (time, f'Leave team {self.get_team_name(old_team)}', self.player))

    def switch(self, time, old_flag, powers, new_team):
        heapq.heappush(
            self.heap,
            (time, f'Switch to team {self.get_team_name(new_team)}',
             self.player))

    def grab(self, time, new_flag, powers, team):
        heapq.heappush(
            self.heap,
            (time, f'Grab {new_flag!s}', self.player))

    def capture(self, time, old_flag, powers, team):
        heapq.heappush(
            self.heap,
            (time, f'Capture {old_flag!s}', self.player))

    def flagless_capture(self, time, flag, powers, team):
        heapq.heappush(
            self.heap,
            (time, 'Capture marsball', self.player))

    def powerup(self, time, flag, power_up, new_powers, team):
        heapq.heappush(
            self.heap,
            (time, f'Power up {power_up!s}', self.player))

    def duplicate_powerup(self, time, flag, powers, team):
        heapq.heappush(
            self.heap,
            (time, f'Grab duplicate powerup', self.player))

    def powerdown(self, time, flag, power_down, new_powers, team):
        heapq.heappush(
            self.heap,
            (time, f'Power down {power_down!s}', self.player))

    def return_(self, time, flag, powers, team):
        heapq.heappush(self.heap, (time, 'Return', self.player))

    def tag(self, time, flag, powers, team):
        heapq.heappush(self.heap, (time, 'Tag', self.player))

    def drop(self, time, old_flag, powers, team):
        heapq.heappush(
            self.heap,
            (time, f'Drop {old_flag!s}', self.player))

    def pop(self, time, powers, team):
        heapq.heappush(self.heap, (time, 'Pop', self.player))

    def start_prevent(self, time, flag, powers, team):
        heapq.heappush(self.heap, (time, 'Start preventing', self.player))

    def stop_prevent(self, time, flag, powers, team):
        heapq.heappush(self.heap, (time, 'Stop preventing', self.player))

    def start_button(self, time, flag, powers, team):
        heapq.heappush(self.heap, (time, 'Start buttoning', self.player))

    def stop_button(self, time, flag, powers, team):
        heapq.heappush(self.heap, (time, 'Stop buttoning', self.player))

    def start_block(self, time, flag, powers, team):
        heapq.heappush(self.heap, (time, 'Start blocking', self.player))

    def stop_block(self, time, flag, powers, team):
        heapq.heappush(self.heap, (time, 'Stop blocking', self.player))

    def end(self, time, flag, powers, team):
        heapq.heappush(self.heap, (time, 'Game ends', self.player))



import datetime
import json

from tagpro_eu.blob import Blob


class JsonObject:
    """
    A data entity to be read from a tagpro.eu JSON file.

    Subclasses of JsonObject should define a __fields__ attribute, which is a
    dict mapping field names (as specified in the JSON format) to a
    constructor of the value. When initializing a JsonObject with some data,
    the values of these fields will be stored to attributes of the resulting
    object.

    If a key in __fields__ has underscores around it, these will be omitted
    when reading from the JSON data, but not for storing the attribute. This
    is useful if you want to define your own attribute with that name.
    """
    def __init__(self, data, strict=False):
        """
        Initialize a JsonObject using a dict of data loaded from a json file.
        When in strict mode, the method will fail when it encounters a missing
        key or wrong data type in the json data. If strict mode is disabled,
        these values will be set to None.

        :param data: dict of loaded json data for this object
        :param strict: whether or not to use strict mode
        :returns: the loaded JsonObject
        :raises KeyError: when strict mode is enabled and a missing key is
        found
        :raises TypeError, ValueError: when strict mode is enabled and an
        element has the wrong data type
        """
        for f, t in self.__fields__.items():
            try:
                value = t(data[f.strip('_')])
            except (KeyError, ValueError, TypeError):
                if strict:
                    raise
                value = None

            if isinstance(value, JsonObject):
                value.__parent__ = self
            elif isinstance(value, list):
                for item in value:
                    item.__parent__ = self

            setattr(self, f, value)

    @classmethod
    def from_string(cls, s, strict=False):
        """
        Load a JsonObject from a json string.

        :param s: the json string to load from
        :param strict: whether or not to use strict mode
        :returns: the loaded JsonObject
        :raises KeyError: when strict mode is disabled and a missing key is
        found
        :raises TypeError: when strict mode is disabled and an element has the
        wrong data type
        """
        return cls(json.loads(s), strict=strict)

    @classmethod
    def from_file(cls, filename, strict=False):
        """
        Load a JsonObject from a json file.

        :param filename: the name of the json file to load from
        :param strict: whether or not to use strict mode
        :returns: the loaded JsonObject
        :raises KeyError: when strict mode is disabled and a missing key is
        found
        :raises TypeError: when strict mode is disabled and an element has the
        wrong data type
        """
        with open(filename) as f:
            return cls(json.load(f), strict=strict)

    def __eq__(self, other):
        for f in self.__fields__.keys():
            if getattr(self, f) != getattr(other, f):
                return False
        return True

    def __repr__(self):
        return 'JsonObject()'

    def to_dict(self):
        """
        Return a dict containing the data stored in this JsonObject. For a
        JsonObject o the following should hold:

            o == o.__class__(o.to_dict())

        :returns: the corresponding dict
        """
        def f(x):
            if isinstance(x, JsonObject):
                return x.to_dict()
            elif isinstance(x, list):
                return list(map(JsonObject.to_dict, x))
            elif isinstance(x, datetime.datetime):
                return int(x.strftime('%s'))
            elif isinstance(x, Blob):
                return x.to_string()
            else:
                return x

        return {field.strip('_'): f(getattr(self, field))
                for field in self.__fields__}

    def to_json(self):
        return json.dumps(self.to_dict())


class ListOf:
    """
    Helper to serve as a constructor for lists in JsonObjects, that
    contain another type to be parsed. Each object in the created list has
    an index attribute, which corresponds to its index in the given list.
    """
    def __init__(self, t):
        """
        :param t: the constructor to use on the list's elements
        """
        self.type = t

    def __call__(self, data):
        out = []

        for i, d in enumerate(data):
            v = self.type(d)
            v.index = i
            out.append(v)

        return out



from enum import IntEnum, IntFlag


# Teams
class Team(IntEnum):
    """
    The team a player can be on. Actual names of teams can be discovered from
    their corresponding Team objects.
    """
    none = 0
    red = 1
    blue = 2

    def __str__(self):
        if self == Team.none:
            return "No team"
        elif self == Team.red:
            return "Red"
        elif self == Team.blue:
            return "Blue"

    def __eq__(self, other):
        if isinstance(other, Team):
            return super().__eq__(other)
        else:
            return other.__eq__(self)


# Flags
class Flag(IntEnum):
    """
    The flag a player can be holding.
    """
    none = 0
    opponent = 1
    opponent_potato = 2
    neutral = 3
    neutral_potato = 4
    temp = 5

    def __str__(self):
        if self == Flag.none:
            return 'No flag'
        elif self == Flag.opponent:
            return 'Opponent flag'
        elif self == Flag.opponent_potato:
            return 'Opponent potato'
        elif self == Flag.neutral:
            return 'Neutral flag'
        elif self == Flag.neutral_potato:
            return 'Neutral potato'
        elif self == Flag.temp:
            return 'Temporary flag'

    def is_potato(self):
        """
        Return whether or not a flag is a potato.

        :return: whether or not the flag is a potato
        """
        return self in (Flag.opponent_potato, Flag.neutral_potato)

    def is_opponent(self):
        """
        Return whether or not a flag is the opponent's.

        :return: whether or not the flag is the opponent's
        """
        return self in (Flag.opponent, Flag.opponent_potato)

    def is_neutral(self):
        """
        Return whether or not a flag is a neutral flag.

        :return: whether or not the flag is a neutral flag
        """
        return self in (Flag.neutral, Flag.neutral_potato)


# Powerups
class Powerup(IntFlag):
    """
    The powerups a player can have. This is a flag enum, so values can be
    combined using | or +.
    """
    none = 0
    juke_juice = 1
    rolling_bomb = 2
    tagpro = 4
    top_speed = 8
    all = 15

    def __str__(self):
        if self == Powerup.none:
            return 'No powerups'
        elif self == Powerup.juke_juice:
            return 'Juke Juice'
        elif self == Powerup.rolling_bomb:
            return 'Rolling Bomb'
        elif self == Powerup.tagpro:
            return 'TagPro'
        elif self == Powerup.top_speed:
            return 'Top Speed'
        else:
            return ', '.join(str(p)
                             for p in Powerup.enumerate()
                             if self & p > 0)

    @classmethod
    def enumerate(cls):
        """
        Yield all four distinct powerups a player can have.

        :returns: the four powerups
        """
        yield from (
            Powerup.juke_juice,
            Powerup.rolling_bomb,
            Powerup.tagpro,
            Powerup.top_speed
        )


# Tiles
class Tile(IntEnum):
    """
    A tile on a map.
    """
    empty = 0

    wall = 10
    wall_ll = 11
    wall_ul = 12
    wall_ur = 13
    wall_lr = 14

    floor = 20

    team_red = 110
    team_blue = 120
    team_neutral = 230

    flag_red = 30
    flag_blue = 40
    flag_neutral = 160

    potato_red = 190
    potato_blue = 200
    potato_neutral = 210

    flag_temp = 161

    boost = 50
    boost_red = 140
    boost_blue = 150

    bomb = 100

    spike = 70

    powerup = 60
    juke_juice = 61
    rolling_bomb = 62
    tagpro = 63
    top_speed = 64

    button = 80

    gate_open = 90
    gate_green = 91
    gate_red = 92
    gate_blue = 93

    protal_entry = 130
    portal_exit = 131

    endzone_red = 170
    endzone_blue = 180

    marsball = 211

    gravity_well = 220


# Flairs
class Flair(IntEnum):
    """
    A flair a player can use.
    """
    none = 0

    leaderboard_day = 1
    leaderboard_week = 2
    leaderboard_month = 3

    winrate_55 = 4
    winrate_65 = 5
    winrate_75 = 6

    donor_10 = 18
    donor_40 = 20
    donor_100 = 21
    donor_200 = 24
    donor_btc = 25

    developer = 19
    community_contributor = 17
    contest_winner = 22
    moderator = 7
    maptester = 8

    kongregate = 23

    birthday_1 = 33
    birthday_2a = 40
    birthday_2b = 41
    birthday_3a = 55
    birthday_3b = 56
    birthday_4a = 72
    birthday_4b = 73
    birthday_4c = 74

    st_patricks_1 = 34
    st_patricks_2 = 42
    st_patricks_3 = 57

    april_1 = 35
    april_2 = 43

    easter_1 = 36
    easter_2a = 49
    easter_2b = 50
    easter_3a = 58
    easter_3b = 59
    easter_3c = 65
    easter_4a = 44
    easter_4b = 45

    unfortunatesniper = 37

    halloween_1a = 38
    halloween_1b = 39
    halloween_2a = 52
    halloween_2b = 53
    halloween_2c = 54
    halloween_3a = 66
    halloween_3b = 67
    halloween_3c = 68
    halloween_4a = 75
    halloween_4b = 76
    halloween_4c = 77

    lgbt = 51

    christmas_1a = 69
    christmas_1b = 70
    christmas_1c = 71

    degree_2 = 116
    degree_6 = 81
    degree_9 = 117
    degree_11 = 82
    degree_17 = 104
    degree_32 = 83
    degree_42 = 84
    degree_51 = 85
    degree_57 = 118
    degree_66 = 86
    degree_69 = 87
    degree_79 = 105
    degree_88 = 88
    degree_98 = 89
    degree_100 = 90
    degree_101 = 97
    degree_110 = 119
    degree_123 = 98
    degree_130 = 106
    degree_143 = 99
    degree_151 = 100
    degree_168 = 101
    degree_180 = 102
    degree_196 = 103
    degree_206 = 121
    degree_212 = 91
    degree_220 = 120
    degree_238 = 107
    degree_276 = 115
    degree_300 = 113
    degree_314 = 114



class Time(int):
    """
    Represents an amount of time in-game.
    Casting to int will give the number of frames.
    """
    @classmethod
    def from_seconds(cls, x):
        """
        Create a Time from a number of seconds.
        """
        return cls(x * 60)

    @classmethod
    def from_minutes(cls, x):
        """
        Create a Time from a number of minutes.
        """
        return cls(x * 3600)

    @property
    def seconds(self):
        """
        The number of seconds in this time, discarding the remainder frames.
        """
        return self // 60

    @property
    def minutes(self):
        """
        The number of minutes in this time, discarding the remainder seconds
        and frames.
        """
        return self // 3600

    def __add__(self, other):
        return Time(super().__add__(other))

    def __sub__(self, other):
        return Time(super().__sub__(other))

    def __str__(self):
        """
        Format a time in in-game frames, to show minutes, seconds and
        centiseconds.

        :param frames: the time in frames
        :returns: the formatted time
        """
        m, s = divmod(self, 3600)
        s /= 60

        return f'{m:0>2}:{s:0>5.2f}'

    def __repr__(self):
        return str(self)



import re
import requests

import tagpro_eu.match


def download_match(url=None, id=None, raw_url=None):
    """
    Download the match with the given ID or URL, and return it as a Match
    object.

    - If id is given, the match will be downloaded from tagpro.eu
    - If url is given, the match ID will be extracted from it
      - If this succeeds, the match with given ID will be downloaded from
        tagpro.eu
      - Otherwise, it will attempt to directly download the data from url,
        or raw_url if it was given
    - If raw_url is given, the data will be directly downloaded from there

    Note that a single match ID as url will also succeed. If you take match
    URLs as user input, you can usually just pass it as url parameter, which
    will work for match links, download links, match IDs and raw data links
    for matches hosted elsewhere.

    This method should probably not be called too many times, as it makes
    requests to tagpro.eu. If you have to download multiple matches, consider
    using a bulk file, which you can (manually) download on
    https://tagpro.eu/?science. The module tagpro_eu.bulk has methods for
    loading those.

    :param id: a match ID
    :param url: a tagpro.eu match URL
    :param raw_url: a direct link to raw data that doesn't have to be on
    tagpro.eu
    :returns: the corresponding Match object
    :raises ValueError: when no valid id, url or raw_url was given
    """
    if id is None and url is not None:
        id = match_url_to_id(url)

        # if match_url_to_id didn't work and raw_url isn't set already,
        # we will use url as raw_url
        if raw_url is None:
            raw_url = url

    if id is not None:
        raw_url = f'https://tagpro.eu/data/?match={id}'

    if raw_url is None:
        raise ValueError("No valid match ID or URL was given")

    r = requests.get(raw_url)
    return tagpro_eu.match.Match(r.json())


match_url_regex =\
    r'^((https?://)?(www\.)?tagpro\.eu/?\?(download|match)=)?(\d+)$'


def match_url_to_id(url):
    """
    Converts a tagpro.eu match URL to the integer ID of that match.
    The URL can be a match page (/?match={id}), a raw data download link
    (/?download={id}), or just an ID.

    :param url: the URL to extract the ID from
    :returns: the ID of the match, or None if the URL didn't match
    """
    match = re.match(match_url_regex, str(url))
    if match is not None:
        return int(match.group(5))
    else:
        return None



from tagpro_eu import Time
import unittest


class TestTime(unittest.TestCase):
    def test_int(self):
        self.assertEqual(Time(0), 0)
        self.assertNotEqual(Time(0), 1)

    def test_arithmetic(self):
        self.assertEqual(Time(1) + Time(2), Time(3))
        self.assertEqual(Time(10) - Time(3), Time(7))

        self.assertIsInstance(Time(1) + Time(2), Time)
        self.assertIsInstance(Time(2) - Time(1), Time)

    def test_constructors(self):
        self.assertEqual(Time(60), Time.from_seconds(1))
        self.assertEqual(Time(7200), Time.from_minutes(2))
        self.assertEqual(Time.from_seconds(180), Time.from_minutes(3))

    def test_properties(self):
        self.assertEqual(Time.from_seconds(10).seconds, 10)
        self.assertEqual(Time.from_minutes(10).minutes, 10)
        self.assertEqual(Time().seconds, 0)
        self.assertEqual(Time().minutes, 0)

    def test_str(self):
        self.assertEqual(str(Time(0)), '00:00.00')
        self.assertEqual(str(Time.from_seconds(30)), '00:30.00')
        self.assertEqual(str(Time.from_minutes(5)), '05:00.00')
        self.assertEqual(str(Time(3900)), '01:05.00')
        self.assertEqual(str(Time(90)), '00:01.50')



from collections import namedtuple
import datetime
import heapq

from tagpro_eu.blob import Blob
from tagpro_eu.constants import Team
from tagpro_eu.data import JsonObject
from tagpro_eu.data import ListOf
from tagpro_eu.map import Map
from tagpro_eu.player import Player
from tagpro_eu.player import PlayerEventHandler
from tagpro_eu.player import PlayerEventLogger
from tagpro_eu.player import PlayerStats
from tagpro_eu.util import Time


class Splat(namedtuple('Splat', ['time', 'x', 'y', 'player', 'team'])):
    __slots__ = ()

    def __lt__(self, other):
        return self.time < other.time


class MatchTeam(JsonObject):
    """
    Represent a team object in tagpro.eu match files.
    The exact format can be found on https://tagpro.eu/?science
    """
    __fields__ = {
        'name': str,
        'score': int,
        '__splats__': Blob.from_b64
    }

    def __init__(self, data, strict=False):
        super().__init__(data, strict=strict)

        self.__splatlist__ = None
        self.__stats__ = None

    @property
    def players(self):
        """
        Return a list of players on this team.
        """
        return list(filter(lambda p: p.team == self, self.__parent__.players))

    @property
    def team(self):
        """
        Return the Team value (from tagpro_eu.constants) for this team.
        """
        return Team(self.index + 1)

    @property
    def splats(self):
        """
        Return a list of splats on the map for this team. This is lazy-loaded
        from the __splats__ blob.

        :returns: list of splats
        """
        if self.__splatlist__ is None:
            self._parse_splats()

        return self.__splatlist__

    @property
    def stats(self):
        """
        Return a PlayerStats object containing the aggregate of all this team's
        player's stats.

        TODO: handle team switches properly
        """
        if self.__stats__ is None:
            self.__stats__ = PlayerStats()
            for player in self.players:
                s = PlayerStats()
                player.parse_events(s)
                self.__stats__ += s
        return self.__stats__

    def _parse_splats(self):
        def bits(size):
            size *= 40
            grid = size - 1
            result = 32
            if not (grid & 0xFFFF0000):
                result -= 16
                grid <<= 16
            if not (grid & 0xFF000000):
                result -= 8
                grid <<= 8
            if not (grid & 0xF0000000):
                result -= 4
                grid <<= 4
            if not (grid & 0xC0000000):
                result -= 2
                grid <<= 2
            if not (grid & 0x80000000):
                result -= 1

            return (result, ((1 << result) - size >> 1) + 20)

        # Determine the players who the splats belong to
        player_splats = []

        class PlayerSplat(namedtuple('PlayerSplat', ['time', 'player'])):
            __slots__ = ()

            def __lt__(self, other):
                return self.time < other.time

        class PlayerSplatsHandler(PlayerEventHandler):
            def __init__(self, player, team):
                self.player = player
                self.team = team

            def drop(self, time, old_flag, powers, team):
                self.pop(time, powers, team)

            def pop(self, time, powers, team):
                if team == self.team:
                    heapq.heappush(player_splats,
                                   PlayerSplat(time, self.player))

        for p in self.__parent__.players:
            h = PlayerSplatsHandler(p, self.team)
            p.parse_events(h)

        blob = self.__splats__
        blob.reset()

        self.__splatlist__ = []

        x = bits(self.__parent__.map.width)
        y = bits(self.__parent__.map.height)

        index = 0

        while not blob.end():
            n = blob.read_tally()

            if n > 0:
                for i in range(n):
                    ps = heapq.heappop(player_splats)
                    self.__splatlist__.append(
                        Splat(ps.time,
                              blob.read_fixed(x[0]) - x[1],
                              blob.read_fixed(y[0]) - y[1],
                              ps.player,
                              self))

            index += 1

    def __eq__(self, other):
        """
        I don't see a reason to compare teams across matches, so comparing
        parents is fine.
        """
        if isinstance(other, Team):
            return self.team == other

        return other is not None and self.index == other.index and\
            self.__parent__ == other.__parent__

    def __repr__(self):
        return f'MatchTeam(name={self.name!r})'


class Match(JsonObject):
    """
    Represents a match object in tagpro.eu match files.
    The exact format can be found on https://tagpro.eu/?science
    """
    __fields__ = {
        'server': str,               # Domain name of the game server
        'port': int,
        'official': bool,            # Played on an "official" server
        'group': str,                # Group ID
        # Start of the match
        'date': datetime.datetime.fromtimestamp,
        'timeLimit': int,            # In minutes
        'duration': Time,            # In frames
        'finished': bool,
        'map': Map,                  # Map object
        'players': ListOf(Player),   # Array of player objects
        'teams': ListOf(MatchTeam),  # Array of team objects
    }

    def __init__(self, data, strict=False):
        super().__init__(data, strict=strict)

        self.__splats__ = None

    def team(self, team):
        """
        Return the MatchTeam object corresponding to the given Team enum value.

        :param team: the team index (from tagpro_eu.constants.Team)
        :returns: the MatchTeam object
        """
        if team == Team.red:
            return self.teams[0]
        elif team == Team.blue:
            return self.teams[1]
        else:
            return None

    @property
    def team_red(self):
        """
        Return the red team.
        """
        return self.team(Team.red)

    @property
    def team_blue(self):
        """
        Return the blue team.
        """
        return self.team(Team.blue)

    @property
    def splats(self):
        """
        Return a list of all splats in the game.
        """
        if self.__splats__ is None:
            sr = self.team_red.splats
            sb = self.team_blue.splats
            self.__splats__ = []

            ir, ib = 0, 0

            while ir < len(sr) or ib < len(sb):
                if ib == len(sb) or ir < len(sr) and sr[ir] < sb[ib]:
                    self.__splats__.append(sr[ir])
                    ir += 1
                else:
                    self.__splats__.append(sb[ib])
                    ib += 1

        return self.__splats__

    def __eq__(self, other):
        return self.server == other.server and\
               self.port == other.port and\
               self.date == other.date

    def __compute_cap_diff__(self):
        """
        Compute the caps_from and caps_against for all players in the game.
        """
        Cap = namedtuple('Cap', ['time', 'team'])
        caps = []

        class CapAggregator(PlayerEventHandler):
            def capture(self, time, _, __, team):
                caps.append(Cap(time, team))

        agg = CapAggregator()

        for player in self.players:
            player.parse_events(agg)

        caps.sort()

        class CapDiffHandler(PlayerEventHandler):
            def __init__(self, caplist, player):
                self.team = Team.none
                self.player = player
                self.caplist = caplist
                self.index = 0

            def catch_up(self, time):
                while self.index < len(self.caplist)\
                        and self.caplist[self.index].time <= time:
                    if self.team != Team.none:
                        capteam = self.caplist[self.index].team
                        if capteam == self.team:
                            self.player.__caps_for__ += 1
                        else:
                            self.player.__caps_against__ += 1
                    self.index += 1

            def join(self, time, team):
                self.catch_up(time)
                self.team = team

            def quit(self, time, _, __, ___):
                self.catch_up(time)
                self.team = Team.none

            def switch(self, time, _, __, team):
                self.catch_up(time)
                self.team = team

            def end(self, time, _, __, team):
                self.catch_up(time)
                self.team = Team.none

        for player in self.players:
            player.__caps_for__ = 0
            player.__caps_against__ = 0
            player.parse_events(CapDiffHandler(caps, player))

    def create_timeline(self, sort=False):
        """
        Return a timeline of events for all players in the match. Each event
        is a tuple of the time, the event (as a string) and the corresponding
        Player object.

        If sort is set to False (default), the timeline will be returned as a
        priority queue (MinHeap). The minimum (that is, occurred first) value
        can be obtained by repeatedly calling heappop from the module heapq.
        Using a priority queue increases performance for creating the timeline.
        If sort is set to True, the timeline is automatically sorted and you
        don't have to bother using heapq.

        :param sort: whether or not to sort the timeline
        :returns: the match timeline
        """
        heap = []

        for player in self.players:
            handler = PlayerEventLogger(heap, player)
            player.parse_events(handler)

        if sort:
            return sorted(heap)
        return heap

    def print_timeline(self):
        """
        Format and print the timeline generated by create_timeline.
        """
        timeline = self.create_timeline()

        while timeline:
            time, event, player = heapq.heappop(timeline)
            print(f'{time} | {player.name:<12} | {event}')

    def __repr__(self):
        return f'Match(server={self.server!r}, port={self.port!r})'

    def print_scoreboard(self, sort_key=None, fields=None):
        """
        Format and print a scoreboard for the match. If desired, a custom
        sorting key and list of fields to show can be given. By default, the
        players will be sorted by their score in descending order, and all
        fields will be shown.

        In fields, every field should be given as a string. The possible fields
        are every attribute of PlayerStats, plus name and score.

        :param sort_key: function that returns the sorting key for a player
        :param fields: the list of fields (as strings) to show
        """
        def default_sort(p):
            return -p.score

        non_stat_attrs = ('name', 'score', 'points')

        def format_field(p, field):
            if field in non_stat_attrs:
                v = getattr(p, field)
            else:
                v = getattr(p.stats, field)

            return str(v).center(field_width(field))

        def field_name(field):
            return 'Powerups' if field == 'pups_total' else field.title()

        def field_width(field):
            return 12 if field == 'name' else 8

        if sort_key is None:
            sort_key = default_sort

        if fields is None:
            fields = ['name', 'score', 'time', 'tags', 'pops', 'grabs',
                      'drops', 'hold', 'captures', 'returns', 'prevent',
                      'button', 'block', 'pups_total', 'points']

        line = '+' + '+'.join('-' * field_width(f) for f in fields) + '+'
        print(line)
        print('|' + '|'.join(field_name(f).center(field_width(f))
                             for f in fields) + '|')
        print(line)

        for p in sorted(self.players, key=sort_key):
            print('|' + '|'.join(format_field(p, f) for f in fields) + '|')

        print(line)



#!/usr/bin/env python3
from setuptools import find_packages, setup

from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_desc = f.read()

setup(
    name='tagpro-eu',
    version='1.1.2',
    license='GPL-3.0',

    packages=find_packages(exclude=['test', 'examples']),

    description='A Python parser for tagpro.eu matches',
    long_description=long_desc,
    url='https://github.com/arfie/tagpro-eu-python',
    author='Ruud Verbeek',

    keywords='tagpro',

    install_requires='requests',
    python_requires='>=3.6',

    test_suite='test',

    py_modules=['tagpro_eu'],

    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',

        'Programming Language :: Python :: 3.6',

        'Topic :: Games/Entertainment',
    ]
)



import base64


class Blob:
    """
    A blob object as loaded from a json object. The exact format can be found
    on https://tagpro.eu/?science.
    """
    def __init__(self, data):
        """
        Initialize a blob with the given bytestring as data.

        :param data: the data for this blob
        """
        self.data = data
        self.pos = 0

    @classmethod
    def from_b64(cls, b64data):
        """
        Initialize a blob with the given base64-encoded data.

        :param data: the base64-encoded data for this blob
        """
        return cls(base64.b64decode(b64data))

    def end(self):
        """
        Return whether or not the pointer has reached the end of the blob.

        :returns: whether or not reading is finished
        """
        return self.pos >> 3 >= len(self.data)

    def read_bool(self):
        """
        Read one boolean from the blob.

        This code is translated from the PHP implementation on
        https://tagpro.eu/?science.

        :returns: the boolean read from the blob
        """
        res = 0
        if not self.end():
            res = self.data[self.pos >> 3] >> 7 - (self.pos & 7) & 1
        self.pos += 1
        return res

    def read_fixed(self, bits):
        """
        Read one n-bit integer from the blob.

        This code is translated from the PHP implementation on
        https://tagpro.eu/?science.

        :param bits: the number of bits the integer fits in
        :returns: the integer read from the blob
        """
        res = 0
        for _ in range(bits):
            res = res << 1 | self.read_bool()
        return res

    def read_tally(self):
        """
        Read a tally (arbitrarily large unsigned integer) from the blob.

        This code is translated from the PHP implementation on
        https://tagpro.eu/?science.

        :returns: the tally read from the blob
        """
        res = 0
        while self.read_bool():
            res += 1
        return res

    def read_footer(self):
        """
        Read a footer (unsigned integer) from the blob.

        This code is translated from the PHP implementation on
        https://tagpro.eu/?science.

        :returns: the footer read from the blob
        """
        size = self.read_fixed(2) << 3
        free = 8 - (self.pos & 7) & 7
        size |= free
        minimum = 0

        while free < size:
            minimum += 1 << free
            free += 8

        return self.read_fixed(size) + minimum

    def reset(self):
        """
        Reset the pointer to the start of the blob.
        """
        self.pos = 0

    def __repr__(self):
        return 'Blob()'

    def to_string(self):
        """
        base64-encode the blob data.

        :returns: the base64-encoded blob data
        """
        return base64.b64encode(self.data).decode('ascii')



from tagpro_eu.blob import Blob
from tagpro_eu.constants import Tile
from tagpro_eu.data import JsonObject


class Map(JsonObject):
    """
    Represents a map object in tagpro.eu match files.
    The exact format can be found on https://tagpro.eu/?science
    """
    __fields__ = {
        'name': str,
        'author': str,
        'type': str,        # ctf, nf, mb etc., empty for unofficial maps
        'marsballs': int,   # number
        'width': int,
        '__tiles__': Blob.from_b64,
    }

    def __init__(self, data, strict=False):
        super().__init__(data, strict=strict)

        self.__tilemap__ = None

    @property
    def tiles(self):
        """
        Return the 2D array of tiles on this map. This is lazy-loaded from
        the blob in __tiles__ and then stored in __tilemap__.

        :returns: the tiles on the map
        """
        if self.__tilemap__ is None:
            self._parse_tiles()

        return self.__tilemap__

    @property
    def height(self):
        """
        Return the height of the map. Unlike width, this is not stored in the
        JSON format, and is therefore lazy-loaded together with tiles.

        :returns: the height of the map in tiles
        """
        if self.__tilemap__ is None:
            self._parse_tiles()

        return len(self.__tilemap__)

    def _parse_tiles(self):
        """
        Load __tilemap__ from the __tiles__ blob, to be used by the tiles and
        height properties.
        """
        self.__tilemap__ = []

        blob = self.__tiles__
        blob.reset()

        x, y = 0, 0

        while not blob.end() or x > 0:
            tile = blob.read_fixed(6)
            if tile != Tile.empty:
                # idk
                if tile < 6:
                    tile += 9
                elif tile < 13:
                    tile = (tile - 4) * 10
                elif tile < 17:
                    tile += 77
                elif tile < 20:
                    tile = (tile - 7) * 10
                elif tile < 22:
                    tile += 110
                else:
                    tile = (tile - 8) * 10

            for i in range(blob.read_footer() + 1):
                if x == 0:
                    self.__tilemap__.append([])
                self.__tilemap__[y].append(Tile(tile))

                x += 1
                if x == self.width:
                    x = 0
                    y += 1

    def __eq__(self, other):
        """
        Equality of maps is determined by comparing tiles.

        Map might have multiple iterations, so just name(+author) won't work.
        Maps might be renamed, but still have the same tiles.
        Maps will probably have to be compared across matches, so comparing
        parents is not a good idea.
        """
        return other is not None and self.tiles == other.tiles

    def __repr__(self):
        return f'Map(name={self.name!r})'
